@@ -1,4 +1,4 @@
-// Copyright 2017 Lucas Morales <lucas@lucasem.com>
+// Copyright 2017-2019 Lucas Morales <lucas@lucasem.com>
 // Derived from work according to the following notices:
 //
 // Copyright 2014 The Rust Project Developers. See the COPYRIGHT
@@ -26,6 +26,9 @@
 //! }
 //! ```
 //!
+//! The [`Pool`] api manages workers and tasks. Enable the `crossbeam` feature if you want to use
+//! crossbeam's channels rather than [`std::sync::mpsc`] for the pool's task management.
+//!
 //! # Examples
 //!
 //! A worker is provided in [`workerpool::thunk`], a stateless [`ThunkWorker<T>`]. It executes on
@@ -34,9 +37,12 @@
 //!
 //! ```
 //! # extern crate workerpool;
+//! # #[cfg(feature = "crossbeam")] extern crate crossbeam_channel;
+//! # #[cfg(feature = "crossbeam")] use crossbeam_channel::unbounded as channel;
+//! # #[cfg(not(feature = "crossbeam"))]
+//! use std::sync::mpsc::channel;
 //! use workerpool::Pool;
 //! use workerpool::thunk::{Thunk, ThunkWorker};
-//! use std::sync::mpsc::channel;
 //!
 //! fn main() {
 //!     let n_workers = 4;
@@ -61,11 +67,14 @@
 //!
 //! ```
 //! # extern crate workerpool;
+//! # #[cfg(feature = "crossbeam")] extern crate crossbeam_channel;
+//! # #[cfg(feature = "crossbeam")] use crossbeam_channel::unbounded as channel;
+//! # #[cfg(not(feature = "crossbeam"))]
+//! use std::sync::mpsc::channel;
 //! use workerpool::{Worker, Pool};
 //! use std::process::{Command, ChildStdin, ChildStdout, Stdio};
 //! use std::io::prelude::*;
 //! use std::io::{self, BufReader};
-//! use std::sync::mpsc::channel;
 //!
 //! struct LineDelimitedProcess {
 //!     stdin: ChildStdin,
@@ -128,13 +137,23 @@
 //! [`ThunkWorker<T>`]: thunk/struct.ThunkWorker.html
 //! [`Thunk::of`]: thunk/struct.Thunk.html#method.of
 //! [`Thunk<T>`]: thunk/struct.Thunk.html
+//! [`Pool`]: struct.Pool.html
+//! [`std::sync::mpsc`]: https://doc.rust-lang.org/stable/std/sync/mpsc/
 
+#[cfg(feature = "crossbeam")]
+extern crate crossbeam_channel;
 extern crate num_cpus;
+extern crate parking_lot;
 
+#[cfg(feature = "crossbeam")]
+use crossbeam_channel::{unbounded as channel, Receiver, Sender};
+#[cfg(not(feature = "crossbeam"))]
+use std::sync::mpsc::{channel, Receiver, Sender};
+
+use parking_lot::{Condvar, Mutex};
 use std::fmt;
-use std::sync::mpsc::{channel, Sender, Receiver};
-use std::sync::{Arc, Mutex, Condvar};
 use std::sync::atomic::{AtomicUsize, Ordering};
+use std::sync::Arc;
 use std::thread;
 
 /// Abstraction of a worker which executes tasks in its own thread.
@@ -172,9 +191,7 @@ where
                 let message = {
                     // Only lock jobs for the time it takes
                     // to get a job, not run it.
-                    let lock = shared_data.job_receiver.lock().expect(
-                        "Worker thread unable to lock job_receiver",
-                    );
+                    let lock = shared_data.job_receiver.lock();
                     lock.recv()
                 };
 
@@ -465,9 +482,7 @@ impl<T: Worker> PoolSharedData<T> {
     /// Notify all observers joining this pool if there is no more work to do.
     fn no_work_notify_all(&self) {
         if !self.has_work() {
-            let _lock = self.empty_trigger.lock().expect(
-                "Unable to notify all joining threads",
-            );
+            let _lock = self.empty_trigger.lock();
             self.empty_condvar.notify_all();
         }
     }
@@ -573,13 +588,14 @@ impl<T: Worker> Pool<T> {
         let job = (inp, None);
         self.jobs
             .lock()
-            .expect("Pool is poisoned")
             .send(job)
             .expect("Pool::execute unable to send job into queue.");
     }
 
     /// Executes with the input on a worker in the pool.
     /// Non-blocking and sends output of the worker's execution to the given sender.
+    /// If you want to use [`crossbeam_channel::Sender`] instead of
+    /// [`std::sync::mpsc::Sender`], enable the `crossbeam` feature for this library.
     ///
     /// # Examples
     ///
@@ -587,9 +603,12 @@ impl<T: Worker> Pool<T> {
     ///
     /// ```
     /// # extern crate workerpool;
+    /// # #[cfg(feature = "crossbeam")] extern crate crossbeam_channel;
+    /// # #[cfg(feature = "crossbeam")] use crossbeam_channel::unbounded as channel;
+    /// # #[cfg(not(feature = "crossbeam"))]
+    /// use std::sync::mpsc::channel;
     /// use workerpool::Pool;
     /// use workerpool::thunk::{Thunk, ThunkWorker};
-    /// use std::sync::mpsc::channel;
     ///
     /// # fn main() {
     /// let pool: Pool<ThunkWorker<i32>> = Pool::new(2);
@@ -601,12 +620,13 @@ impl<T: Worker> Pool<T> {
     /// assert_eq!(10, rx.iter().take(4).sum());
     /// # }
     /// ```
+    /// [`crossbeam_channel::Sender`]: http://docs.rs/crossbeam-channel/*/crossbeam_channel/struct.Sender.html
+    /// [`std::sync::mpsc::Sender`]: https://doc.rust-lang.org/stable/std/sync/mpsc/struct.Sender.html
     pub fn execute_to(&self, tx: Sender<T::Output>, inp: T::Input) {
         self.shared_data.queued_count.fetch_add(1, Ordering::SeqCst);
         let job = (inp, Some(tx));
         self.jobs
             .lock()
-            .expect("Pool is poisoned")
             .send(job)
             .expect("Pool::execute unable to send job into queue.");
     }
@@ -734,7 +754,7 @@ impl<T: Worker> Pool<T> {
     /// # fn main() {
     /// let mut pool: Pool<ThunkWorker<()>> = Pool::new(4);
     /// for _ in 0..10 {
-    ///     pool.execute(Thunk::of(move || {
+    ///     pool.execute(Thunk::of(|| {
     ///         sleep(Duration::from_secs(100));
     ///     }));
     /// }
@@ -760,10 +780,10 @@ impl<T: Worker> Pool<T> {
     /// ```
     pub fn set_num_threads(&mut self, num_threads: usize) {
         assert!(num_threads >= 1);
-        let prev_num_threads = self.shared_data.max_thread_count.swap(
-            num_threads,
-            Ordering::Release,
-        );
+        let prev_num_threads = self
+            .shared_data
+            .max_thread_count
+            .swap(num_threads, Ordering::Release);
         if let Some(num_spawn) = num_threads.checked_sub(prev_num_threads) {
             // Spawn new threads
             for _ in 0..num_spawn {
@@ -814,12 +834,12 @@ impl<T: Worker> Pool<T> {
         }
 
         let generation = self.shared_data.join_generation.load(Ordering::SeqCst);
-        let mut lock = self.shared_data.empty_trigger.lock().unwrap();
+        let mut lock = self.shared_data.empty_trigger.lock();
 
-        while generation == self.shared_data.join_generation.load(Ordering::Relaxed) &&
-            self.shared_data.has_work()
+        while generation == self.shared_data.join_generation.load(Ordering::Relaxed)
+            && self.shared_data.has_work()
         {
-            lock = self.shared_data.empty_condvar.wait(lock).unwrap();
+            self.shared_data.empty_condvar.wait(&mut lock);
         }
 
         // increase generation if we are the first thread to come out of the loop
@@ -880,7 +900,6 @@ impl<T: Worker> Clone for Pool<T> {
     }
 }
 
-
 /// Create a thread pool with one thread per CPU.
 /// On machines with hyperthreading,
 /// this will create one thread per hyperthread.
@@ -890,7 +909,6 @@ impl<T: Worker> Default for Pool<T> {
     }
 }
 
-
 impl<T: Worker> fmt::Debug for Pool<T> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         f.debug_struct("Pool")
@@ -938,9 +956,12 @@ pub mod thunk {
     //!
     //! ```
     //! # extern crate workerpool;
+    //! # #[cfg(feature = "crossbeam")] extern crate crossbeam_channel;
+    //! # #[cfg(feature = "crossbeam")] use crossbeam_channel::unbounded as channel;
+    //! # #[cfg(not(feature = "crossbeam"))]
+    //! use std::sync::mpsc::channel;
     //! use workerpool::Pool;
     //! use workerpool::thunk::{Thunk, ThunkWorker};
-    //! use std::sync::mpsc::channel;
     //!
     //! # fn main() {
     //! let n_workers = 4;
@@ -1031,11 +1052,11 @@ pub mod thunk {
 
 #[cfg(test)]
 mod test {
-    use super::{Pool, Builder};
     use super::thunk::{Thunk, ThunkWorker};
-    use std::sync::{Arc, Barrier};
+    use super::{Builder, Pool};
     use std::sync::atomic::{AtomicUsize, Ordering};
-    use std::sync::mpsc::{sync_channel, channel};
+    use std::sync::mpsc::{channel, sync_channel};
+    use std::sync::{Arc, Barrier};
     use std::thread::{self, sleep};
     use std::time::Duration;
 
@@ -1046,7 +1067,7 @@ mod test {
         let new_thread_amount = TEST_TASKS + 8;
         let mut pool = Pool::<ThunkWorker<()>>::new(TEST_TASKS);
         for _ in 0..TEST_TASKS {
-            pool.execute(Thunk::of(move || sleep(Duration::from_secs(23))));
+            pool.execute(Thunk::of(|| sleep(Duration::from_secs(23))));
         }
         sleep(Duration::from_secs(1));
         assert_eq!(pool.active_count(), TEST_TASKS);
@@ -1054,7 +1075,7 @@ mod test {
         pool.set_num_threads(new_thread_amount);
 
         for _ in 0..(new_thread_amount - TEST_TASKS) {
-            pool.execute(Thunk::of(move || sleep(Duration::from_secs(23))));
+            pool.execute(Thunk::of(|| sleep(Duration::from_secs(23))));
         }
         sleep(Duration::from_secs(1));
         assert_eq!(pool.active_count(), new_thread_amount);
@@ -1067,11 +1088,11 @@ mod test {
         let new_thread_amount = 2;
         let mut pool = Pool::<ThunkWorker<()>>::new(TEST_TASKS);
         for _ in 0..TEST_TASKS {
-            pool.execute(Thunk::of(move || { 1 + 1; }));
+            pool.execute(Thunk::of(|| ()));
         }
         pool.set_num_threads(new_thread_amount);
         for _ in 0..new_thread_amount {
-            pool.execute(Thunk::of(move || sleep(Duration::from_secs(23))));
+            pool.execute(Thunk::of(|| sleep(Duration::from_secs(23))));
         }
         sleep(Duration::from_secs(1));
         assert_eq!(pool.active_count(), new_thread_amount);
@@ -1083,7 +1104,7 @@ mod test {
     fn test_active_count() {
         let pool = Pool::<ThunkWorker<()>>::new(TEST_TASKS);
         for _ in 0..2 * TEST_TASKS {
-            pool.execute(Thunk::of(move || loop {
+            pool.execute(Thunk::of(|| loop {
                 sleep(Duration::from_secs(10))
             }));
         }
@@ -1101,7 +1122,9 @@ mod test {
         let (tx, rx) = channel();
         for _ in 0..TEST_TASKS {
             let tx = tx.clone();
-            pool.execute(Thunk::of(move || { tx.send(1).unwrap(); }));
+            pool.execute(Thunk::of(move || {
+                tx.send(1).unwrap();
+            }));
         }
 
         assert_eq!(rx.iter().take(TEST_TASKS).fold(0, |a, b| a + b), TEST_TASKS);
@@ -1119,7 +1142,7 @@ mod test {
 
         // Panic all the existing threads.
         for _ in 0..TEST_TASKS {
-            pool.execute(Thunk::of(move || panic!("Ignore this panic, it must!")));
+            pool.execute(Thunk::of(|| panic!("intentional panic")));
         }
         pool.join();
 
@@ -1129,7 +1152,9 @@ mod test {
         let (tx, rx) = channel();
         for _ in 0..TEST_TASKS {
             let tx = tx.clone();
-            pool.execute(Thunk::of(move || { tx.send(1).unwrap(); }));
+            pool.execute(Thunk::of(move || {
+                tx.send(1).unwrap();
+            }));
         }
 
         assert_eq!(rx.iter().take(TEST_TASKS).fold(0, |a, b| a + b), TEST_TASKS);
@@ -1137,7 +1162,6 @@ mod test {
 
     #[test]
     fn test_should_not_panic_on_drop_if_subtasks_panic_after_drop() {
-
         let pool = Pool::<ThunkWorker<()>>::new(TEST_TASKS);
         let waiter = Arc::new(Barrier::new(TEST_TASKS + 1));
 
@@ -1146,7 +1170,7 @@ mod test {
             let waiter = waiter.clone();
             pool.execute(Thunk::of(move || {
                 waiter.wait();
-                panic!("Ignore this panic, it should!");
+                panic!("intentional panic");
             }));
         }
 
@@ -1171,7 +1195,6 @@ mod test {
             let (b0, b1) = (b0.clone(), b1.clone());
 
             pool.execute(Thunk::of(move || {
-
                 // Wait until the pool has been filled once.
                 if i < TEST_TASKS {
                     b0.wait();
@@ -1231,7 +1254,6 @@ mod test {
         assert_eq!(pool.active_count(), test_tasks_begin);
         b1.wait();
 
-
         b2.wait();
         assert_eq!(pool.active_count(), TEST_TASKS);
         b3.wait();
@@ -1289,7 +1311,7 @@ mod test {
         );
 
         let pool = Pool::<ThunkWorker<()>>::new(4);
-        pool.execute(Thunk::of(move || sleep(Duration::from_secs(5))));
+        pool.execute(Thunk::of(|| sleep(Duration::from_secs(5))));
         sleep(Duration::from_secs(1));
         let debug = format!("{:?}", pool);
         assert_eq!(
@@ -1408,7 +1430,9 @@ mod test {
 
         // This batch of jobs will occupy the pool for some time
         for _ in 0..6 {
-            pool.execute(Thunk::of(move || { sleep(Duration::from_secs(2)); }));
+            pool.execute(Thunk::of(|| {
+                sleep(Duration::from_secs(2));
+            }));
         }
 
         // The following jobs will be inserted into the pool in a random fashion
@@ -1421,15 +1445,13 @@ mod test {
                 let (tx, rx) = channel();
                 for i in 0..42 {
                     let tx = tx.clone();
-                    pool.execute(Thunk::of(
-                        move || { tx.send(i).expect("channel will be waiting"); },
-                    ));
+                    pool.execute(Thunk::of(move || {
+                        tx.send(i).expect("channel will be waiting");
+                    }));
                 }
                 drop(tx);
-                rx.iter().fold(
-                    0,
-                    |accumulator, element| accumulator + element,
-                )
+                rx.iter()
+                    .fold(0, |accumulator, element| accumulator + element)
             })
         };
         let t1 = {
@@ -1441,29 +1463,25 @@ mod test {
                 let (tx, rx) = channel();
                 for i in 1..12 {
                     let tx = tx.clone();
-                    pool.execute(Thunk::of(
-                        move || { tx.send(i).expect("channel will be waiting"); },
-                    ));
+                    pool.execute(Thunk::of(move || {
+                        tx.send(i).expect("channel will be waiting");
+                    }));
                 }
                 drop(tx);
-                rx.iter().fold(
-                    1,
-                    |accumulator, element| accumulator * element,
-                )
+                rx.iter()
+                    .fold(1, |accumulator, element| accumulator * element)
             })
         };
 
         assert_eq!(
             861,
-            t0.join().expect(
-                "thread 0 will return after calculating additions",
-            )
+            t0.join()
+                .expect("thread 0 will return after calculating additions",)
         );
         assert_eq!(
             39916800,
-            t1.join().expect(
-                "thread 1 will return after calculating multiplications",
-            )
+            t1.join()
+                .expect("thread 1 will return after calculating multiplications",)
         );
     }
 
@@ -1505,10 +1523,9 @@ mod test {
         let n_cycles = 4;
         let n_workers = 4;
         let (tx, rx) = channel();
-        let builder = Builder::new().num_threads(n_workers).thread_name(
-            "join wavesurfer"
-                .into(),
-        );
+        let builder = Builder::new()
+            .num_threads(n_workers)
+            .thread_name("join wavesurfer".into());
         let p_waiter = builder.clone().build::<ThunkWorker<()>>();
         let p_clock = builder.build::<ThunkWorker<()>>();
 
